@@ -6,22 +6,25 @@ import os
 import shlex
 import shutil
 from gettext import gettext as _
+from pathlib import Path
 
 from gi.repository import GLib
 
 from lutris import runtime
 from lutris.cache import get_cache_path
 from lutris.command import MonitoredCommand
+from lutris.database.games import get_game_by_field
+from lutris.exceptions import UnavailableRunnerError
+from lutris.game import Game
 from lutris.installer.errors import ScriptingError
 from lutris.runners import import_task
-from lutris.util import disks, extract, selective_merge, system
+from lutris.util import extract, linux, selective_merge, system
 from lutris.util.fileio import EvilConfigParser, MultiOrderedDict
 from lutris.util.log import logger
 from lutris.util.wine.wine import WINE_DEFAULT_ARCH, get_wine_version_exe
 
 
 class CommandsMixin:
-
     """The directives for the `installer:` part of the install script."""
 
     def __init__(self):
@@ -30,11 +33,20 @@ class CommandsMixin:
 
     def _get_runner_version(self):
         """Return the version of the runner used for the installer"""
-        if (
-                self.installer.runner in ("wine", "winesteam")
-                and self.installer.script.get(self.installer.runner)
-        ):
-            return self.installer.script[self.installer.runner].get("version")
+        if self.installer.runner == "wine":
+            # If a version is specified in the script choose this one
+            if self.installer.script.get(self.installer.runner):
+                return self.installer.script[self.installer.runner].get("version")
+            # If the installer is a extension, use the wine version from the base game
+            if self.installer.requires:
+                db_game = get_game_by_field(self.installer.requires, field="installer_slug")
+                if not db_game:
+                    db_game = get_game_by_field(self.installer.requires, field="slug")
+                if not db_game:
+                    logger.warning("Can't find game %s", self.installer.requires)
+                    return None
+                game = Game(db_game["id"])
+                return game.config.runner_config["version"]
         if self.installer.runner == "libretro":
             return self.installer.script["game"]["core"]
         return None
@@ -52,13 +64,15 @@ class CommandsMixin:
                         param_present = True
                 if not param_present:
                     raise ScriptingError(
-                        "One of %s parameter is mandatory for the %s command" % (" or ".join(param), command_name),
+                        _("One of {params} parameter is mandatory for the {cmd} command").format(
+                            params=_(" or ").join(param), cmd=command_name),
                         command_data,
                     )
             else:
                 if param not in command_data:
                     raise ScriptingError(
-                        "The %s parameter is mandatory for the %s command" % (param, command_name),
+                        _("The {param} parameter is mandatory for the {cmd} command").format(
+                            param=param, cmd=command_name),
                         command_data,
                     )
 
@@ -74,7 +88,7 @@ class CommandsMixin:
         """Make filename executable"""
         filename = self._substitute(filename)
         if not system.path_exists(filename):
-            raise ScriptingError("Invalid file '%s'. Can't make it executable" % filename)
+            raise ScriptingError(_("Invalid file '%s'. Can't make it executable") % filename)
         system.make_executable(filename)
 
     def execute(self, data):
@@ -87,11 +101,18 @@ class CommandsMixin:
             self._check_required_params([("file", "command")], data, "execute")
             if "command" in data and "file" in data:
                 raise ScriptingError(
-                    "Parameters file and command can't be used "
-                    "at the same time for the execute command",
+                    _("Parameters file and command can't be used "
+                      "at the same time for the execute command"),
                     data,
                 )
-            file_ref = data.get("file", "")
+
+            # Accept return codes other than 0
+            if "return_code" in data:
+                return_code = data.pop("return_code")
+            else:
+                return_code = "0"
+
+            exec_path = data.get("file", "")
             command = data.get("command", "")
             args_string = data.get("args", "")
             for arg in shlex.split(args_string):
@@ -115,36 +136,37 @@ class CommandsMixin:
             include_processes = []
             exclude_processes = []
         else:
-            raise ScriptingError("No parameters supplied to execute command.", data)
+            raise ScriptingError(_("No parameters supplied to execute command."), data)
 
         if command:
-            file_ref = "bash"
-            args = ["-c", self._get_file(command.strip())]
+            exec_path = "bash"
+            args = ["-c", self._get_file_path(command.strip())]
             include_processes.append("bash")
         else:
             # Determine whether 'file' value is a file id or a path
-            file_ref = self._get_file(file_ref)
-        if system.path_exists(file_ref) and not system.is_executable(file_ref):
-            logger.warning("Making %s executable", file_ref)
-            system.make_executable(file_ref)
-        exec_path = system.find_executable(file_ref)
-        if not exec_path:
-            raise ScriptingError("Unable to find executable %s" % file_ref)
+            exec_path = self._get_file_path(exec_path)
+        if system.path_exists(exec_path) and not system.is_executable(exec_path):
+            logger.warning("Making %s executable", exec_path)
+            system.make_executable(exec_path)
+        exec_abs_path = system.find_executable(exec_path)
+        if not exec_abs_path:
+            raise ScriptingError(_("Unable to find executable %s") % exec_path)
 
         if terminal:
-            terminal = system.get_default_terminal()
+            terminal = linux.get_default_terminal()
 
         if not working_dir or not os.path.exists(working_dir):
             working_dir = self.target_path
 
         command = MonitoredCommand(
-            [exec_path] + args,
+            [exec_abs_path] + args,
             env=env,
             term=terminal,
             cwd=working_dir,
             include_processes=include_processes,
             exclude_processes=exclude_processes,
         )
+        command.accepted_return_code = return_code
         command.start()
         GLib.idle_add(self.parent.attach_logger, command)
         self.heartbeat = GLib.timeout_add(1000, self._monitor_task, command)
@@ -154,12 +176,15 @@ class CommandsMixin:
         """Extract a file, guessing the compression method."""
         self._check_required_params([("file", "src")], data, "extract")
         src_param = data.get("file") or data.get("src")
-        filespec = self._get_file(src_param)
+        filespec = self._get_file_path(src_param)
 
-        filenames = glob.glob(filespec)
+        if os.path.exists(filespec):
+            filenames = [filespec]
+        else:
+            filenames = glob.glob(filespec)
 
         if not filenames:
-            raise ScriptingError("%s does not exist" % filespec)
+            raise ScriptingError(_("%s does not exist") % filespec)
         if "dst" in data:
             dest_path = self._substitute(data["dst"])
         else:
@@ -224,11 +249,11 @@ class CommandsMixin:
         if extra_path:
             drives = [extra_path]
         else:
-            drives = disks.get_mounted_discs()
+            drives = system.get_mounted_discs()
         for drive in drives:
             required_abspath = os.path.join(drive, requires)
             required_abspath = system.fix_path_case(required_abspath)
-            if required_abspath:
+            if required_abspath and system.path_exists(required_abspath):
                 logger.debug("Found %s on cdrom %s", requires, drive)
                 self.game_disc = drive
                 self._iter_commands()
@@ -253,9 +278,8 @@ class CommandsMixin:
             if params.get("optional"):
                 logger.info("Optional path %s not present", src)
                 return
-            raise ScriptingError("Source does not exist: %s" % src, params)
-        if not os.path.exists(dst):
-            os.makedirs(dst)
+            raise ScriptingError(_("Source does not exist: %s") % src, params)
+        os.makedirs(dst, exist_ok=True)
         if os.path.isfile(src):
             # If single file, copy it and change reference in game file so it
             # can be used as executable. Skip copying if the source is the same
@@ -280,7 +304,7 @@ class CommandsMixin:
             if params.get("optional"):
                 logger.info("Optional path %s not present", src)
                 return
-            raise ScriptingError("I can't move %s, it does not exist" % src)
+            raise ScriptingError(_("Invalid source for 'move' operation: %s") % src)
 
         if os.path.isfile(src):
             if os.path.dirname(src) == dst:
@@ -298,22 +322,22 @@ class CommandsMixin:
             else:
                 action = shutil.move
             self._killable_process(action, src, dst)
-        except shutil.Error:
-            raise ScriptingError("Can't move %s \nto destination %s" % (src, dst))
+        except shutil.Error as err:
+            raise ScriptingError(_("Can't move {src} \nto destination {dst}").format(src=src, dst=dst)) from err
 
     def rename(self, params):
         """Rename file or folder."""
         self._check_required_params(["src", "dst"], params, "rename")
         src, dst = self._get_move_paths(params)
         if not os.path.exists(src):
-            raise ScriptingError("Rename error, source path does not exist: %s" % src)
+            raise ScriptingError(_("Rename error, source path does not exist: %s") % src)
         if os.path.isdir(dst):
             try:
                 os.rmdir(dst)  # Remove if empty
             except OSError:
                 pass
         if os.path.exists(dst):
-            raise ScriptingError("Rename error, destination already exists: %s" % src)
+            raise ScriptingError(_("Rename error, destination already exists: %s") % src)
         dst_dir = os.path.dirname(dst)
 
         # Pre-move on dest filesystem to avoid error with
@@ -328,15 +352,15 @@ class CommandsMixin:
         """Process raw 'src' and 'dst' data."""
         try:
             src_ref = params["src"]
-        except KeyError:
-            raise ScriptingError("Missing parameter src")
+        except KeyError as err:
+            raise ScriptingError(_("Missing parameter src")) from err
         src = self.game_files.get(src_ref) or self._substitute(src_ref)
         if not src:
-            raise ScriptingError("Wrong value for 'src' param", src_ref)
+            raise ScriptingError(_("Wrong value for 'src' param"), src_ref)
         dst_ref = params["dst"]
         dst = self._substitute(dst_ref)
         if not dst:
-            raise ScriptingError("Wrong value for 'dst' param", dst_ref)
+            raise ScriptingError(_("Wrong value for 'dst' param"), dst_ref)
         return src.rstrip("/"), dst.rstrip("/")
 
     def substitute_vars(self, data):
@@ -345,8 +369,8 @@ class CommandsMixin:
         filename = self._substitute(data["file"])
         logger.debug("Substituting variables for file %s", filename)
         tmp_filename = filename + ".tmp"
-        with open(filename, "r") as source_file:
-            with open(tmp_filename, "w") as dest_file:
+        with open(filename, "r", encoding='utf-8') as source_file:
+            with open(tmp_filename, "w", encoding='utf-8') as dest_file:
                 line = "."
                 while line:
                     line = source_file.readline()
@@ -363,6 +387,14 @@ class CommandsMixin:
             runner_name = self.installer.runner
         return runner_name, task_name
 
+    def get_wine_path(self):
+        """Return absolute path of wine version used during the install, but
+        None if the wine exe can't be located."""
+        try:
+            return get_wine_version_exe(self._get_runner_version())
+        except UnavailableRunnerError:
+            return None
+
     def task(self, data):
         """Directive triggering another function specific to a runner.
 
@@ -374,19 +406,23 @@ class CommandsMixin:
             GLib.idle_add(self.parent.cancel_button.set_sensitive, False)
         runner_name, task_name = self._get_task_runner_and_name(data.pop("name"))
 
-        wine_version = None
+        # Accept return codes other than 0
+        if "return_code" in data:
+            return_code = data.pop("return_code")
+        else:
+            return_code = "0"
 
         if runner_name.startswith("wine"):
-            wine_version = self._get_runner_version()
-            if wine_version:
-                data["wine_path"] = get_wine_version_exe(wine_version)
+            wine_path = self.get_wine_path()
+            if wine_path:
+                data["wine_path"] = wine_path
             data["prefix"] = data.get("prefix") \
                 or self.installer.script.get("game", {}).get("prefix") \
                 or "$GAMEDIR"
             data["arch"] = data.get("arch") \
                 or self.installer.script.get("game", {}).get("arch") \
                 or WINE_DEFAULT_ARCH
-            if task_name == "wineexec" and self.script_env:
+            if task_name == "wineexec":
                 data["env"] = self.script_env
 
         for key in data:
@@ -402,17 +438,22 @@ class CommandsMixin:
             data[key] = value
 
         task = import_task(runner_name, task_name)
-        thread = task(**data)
+        command = task(**data)
+        if command:
+            command.accepted_return_code = return_code
         GLib.idle_add(self.parent.cancel_button.set_sensitive, True)
-        if isinstance(thread, MonitoredCommand):
+        if isinstance(command, MonitoredCommand):
             # Monitor thread and continue when task has executed
-            GLib.idle_add(self.parent.attach_logger, thread)
-            self.heartbeat = GLib.timeout_add(1000, self._monitor_task, thread)
+            GLib.idle_add(self.parent.attach_logger, command)
+            self.heartbeat = GLib.timeout_add(1000, self._monitor_task, command)
             return "STOP"
         return None
 
-    def _monitor_task(self, thread):
-        if not thread.is_running:
+    def _monitor_task(self, command):
+        if not command.is_running:
+            logger.debug("Return code: %s", command.return_code)
+            if command.return_code not in (str(command.accepted_return_code), "0"):
+                raise ScriptingError(_("Command exited with code %s") % command.return_code)
             self._iter_commands()
             return False
         return True
@@ -422,18 +463,17 @@ class CommandsMixin:
         self._check_required_params(["file", "content"], params, "write_file")
 
         # Get file
-        dest_file_path = self._get_file(params["file"])
+        dest_file_path = self._get_file_path(params["file"])
 
         # Create dir if necessary
         basedir = os.path.dirname(dest_file_path)
-        if not os.path.exists(basedir):
-            os.makedirs(basedir)
+        os.makedirs(basedir, exist_ok=True)
 
         mode = params.get("mode", "w")
         if not mode.startswith(("a", "w")):
-            raise ScriptingError("Wrong value for write_file mode: '%s'" % mode)
+            raise ScriptingError(_("Wrong value for write_file mode: '%s'") % mode)
 
-        with open(dest_file_path, mode) as dest_file:
+        with open(dest_file_path, mode, encoding='utf-8') as dest_file:
             dest_file.write(self._substitute(params["content"]))
 
     def write_json(self, params):
@@ -441,21 +481,18 @@ class CommandsMixin:
         self._check_required_params(["file", "data"], params, "write_json")
 
         # Get file
-        filename = self._get_file(params["file"])
+        filename = self._get_file_path(params["file"])
 
         # Create dir if necessary
         basedir = os.path.dirname(filename)
-        if not os.path.exists(basedir):
-            os.makedirs(basedir)
+        os.makedirs(basedir, exist_ok=True)
 
         merge = params.get("merge", True)
 
-        if not os.path.exists(filename):
-            # create an empty file
-            with open(filename, "a+"):
-                pass
+        # create an empty file if it doesn't exist
+        Path(filename).touch(exist_ok=True)
 
-        with open(filename, "r+" if merge else "w") as json_file:
+        with open(filename, "r+" if merge else "w", encoding='utf-8') as json_file:
             json_data = {}
             if merge:
                 try:
@@ -474,12 +511,11 @@ class CommandsMixin:
         else:
             self._check_required_params(["file", "section", "key", "value"], params, "write_config")
         # Get file
-        config_file_path = self._get_file(params["file"])
+        config_file_path = self._get_file_path(params["file"])
 
         # Create dir if necessary
         basedir = os.path.dirname(config_file_path)
-        if not os.path.exists(basedir):
-            os.makedirs(basedir)
+        os.makedirs(basedir, exist_ok=True)
 
         merge = params.get("merge", True)
 
@@ -505,7 +541,7 @@ class CommandsMixin:
         with open(config_file_path, "wb") as config_file:
             parser.write(config_file)
 
-    def _get_file(self, fileid):
+    def _get_file_path(self, fileid):
         file_path = self.game_files.get(fileid)
         if not file_path:
             file_path = self._substitute(fileid)
@@ -513,10 +549,97 @@ class CommandsMixin:
 
     def _killable_process(self, func, *args, **kwargs):
         """Run function `func` in a separate, killable process."""
-        process = multiprocessing.Pool(1)
-        result_obj = process.apply_async(func, args, kwargs)
-        self.abort_current_task = process.terminate
-        result = result_obj.get()  # Wait process end & reraise exceptions
-        self.abort_current_task = None
-        logger.debug("Process %s returned: %s", func, result)
-        return result
+        with multiprocessing.Pool(1) as process:
+            result_obj = process.apply_async(func, args, kwargs)
+            self.abort_current_task = process.terminate
+            result = result_obj.get()  # Wait process end & re-raise exceptions
+            self.abort_current_task = None
+            logger.debug("Process %s returned: %s", func, result)
+            return result
+
+    def _extract_gog_game(self, file_id):
+        self.extract({
+            "src": file_id,
+            "dst": "$GAMEDIR",
+            "extractor": "innoextract"
+        })
+        app_path = os.path.join(self.target_path, "app")
+        if system.path_exists(app_path):
+            for app_content in os.listdir(app_path):
+                source_path = os.path.join(app_path, app_content)
+                if os.path.exists(os.path.join(self.target_path, app_content)):
+                    self.merge({"src": source_path, "dst": self.target_path})
+                else:
+                    self.move({"src": source_path, "dst": self.target_path})
+        support_path = os.path.join(self.target_path, "__support/app")
+        if system.path_exists(support_path):
+            self.merge({"src": support_path, "dst": self.target_path})
+
+    def _get_scummvm_arguments(self, gog_config_path):
+        """Return a ScummVM configuration from the GOG config files"""
+        with open(gog_config_path, encoding='utf-8') as gog_config_file:
+            gog_config = json.loads(gog_config_file.read())
+        game_tasks = [task for task in gog_config["playTasks"] if task["category"] == "game"]
+        arguments = game_tasks[0]["arguments"]
+        game_id = arguments.split()[-1]
+        arguments = " ".join(arguments.split()[:-1])
+        base_dir = os.path.dirname(gog_config_path)
+        return {
+            "game_id": game_id,
+            "path": base_dir,
+            "arguments": arguments
+        }
+
+    def autosetup_gog_game(self, file_id, silent=False):
+        """Automatically guess the best way to install a GOG game by inspecting its contents.
+        This chooses the right runner (DOSBox, Wine) for Windows game files.
+        Linux setup files don't use innosetup, they can be unzipped instead.
+        """
+        file_path = self.game_files[file_id]
+        file_list = extract.get_innoextract_list(file_path)
+        dosbox_found = False
+        scummvm_found = False
+        windows_override_found = False  # DOS games that also have a Windows executable
+        for filename in file_list:
+            if "dosbox/dosbox.exe" in filename.lower():
+                dosbox_found = True
+            if "scummvm/scummvm.exe" in filename.lower():
+                scummvm_found = True
+            if "_some_windows.exe" in filename.lower():
+                # There's not a good way to handle exceptions without extracting the .info file
+                # before extracting the game. Added for Quake but GlQuake.exe doesn't run on modern wine
+                windows_override_found = True
+        if dosbox_found and not windows_override_found:
+            self._extract_gog_game(file_id)
+            dosbox_config = {
+                "working_dir": "$GAMEDIR/DOSBOX",
+            }
+            for filename in os.listdir(self.target_path):
+                if filename.endswith("_single.conf"):
+                    dosbox_config["main_file"] = filename
+                elif filename.endswith(".conf"):
+                    dosbox_config["config_file"] = filename
+            self.installer.script["game"] = dosbox_config
+            self.installer.runner = "dosbox"
+        elif scummvm_found:
+            self._extract_gog_game(file_id)
+            arguments = None
+            for filename in os.listdir(self.target_path):
+                if filename.startswith("goggame") and filename.endswith(".info"):
+                    arguments = self._get_scummvm_arguments(os.path.join(self.target_path, filename))
+            if not arguments:
+                raise RuntimeError("Unable to get ScummVM arguments")
+            logger.info("ScummVM config: %s", arguments)
+            self.installer.script["game"] = arguments
+            self.installer.runner = "scummvm"
+        else:
+            args = "/SP- /NOCANCEL"
+            if silent:
+                args += " /SUPPRESSMSGBOXES /VERYSILENT /NOGUI"
+            self.installer.is_gog = True
+            return self.task({
+                "name": "wineexec",
+                "prefix": "$GAMEDIR",
+                "executable": file_id,
+                "args": args
+            })
